from typing import Optional

from kama_sdk.core.core.types import PromVector
from kama_sdk.model.supplier.base.supplier import Supplier


class PromVectorToFirstValueSupplier(Supplier):

  def source_data(self) -> Optional[PromVector]:
    return super(PromVectorToFirstValueSupplier, self).source_data()

  def _compute(self):
    if vector := self.source_data():
      return float(vector[0]['value'][1])



from typing import TypeVar

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import Reconstructor
from kama_sdk.model.base.model import Model
from kama_sdk.model.concern.concern import Concern
from kama_sdk.model.concern import concern as concern_module

T = TypeVar('T', bound='ConcernViewAdapter')


class ConcernViewAdapter(Model):

  @cached_property
  def concern_shell(self) -> Concern:
    return self.inflate_child(
      Concern,
      prop=CONCERN_SHELL_KEY
    )

  @classmethod
  def reconstruct(cls, reconstructor: Reconstructor):
    adapter: T = cls.inflate(reconstructor['adapter_ref'])
    concern = adapter.concern_shell.patch({
      concern_module.SEED_KEY: reconstructor['seed']
    })
    return adapter.patch({'concern': concern})


CONCERN_SHELL_KEY = 'concern_shell'



from flask import Blueprint, jsonify

from kama_sdk.controllers.ctrl_utils import jparse
from kama_sdk.core.core import job_client, utils
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.ktea.ktea_provider import ktea_client
from kama_sdk.serializers import manifest_vars_serial
from kama_sdk.model.variable.manifest_variable import ManifestVariable

BASE = '/api/manifest-variable'

controller = Blueprint('manifest_variables_controller', __name__)


@controller.route(BASE)
def manifest_variables_index():
  """
  Inflates and serializes the current list of chart variable.
  :return: serialized list of chart variable.
  """
  config_man.invalidate_cmap()
  manifest_var_models = ManifestVariable.all_vars()
  serialize = lambda cv: manifest_vars_serial.standard(cv=cv, reload=False)
  serialized = list(map(serialize, manifest_var_models))
  return jsonify(data=serialized)


@controller.route(f"{BASE}/defaults")
def manifest_variables_defaults():
  as_dict = config_man.default_vars()
  return jsonify(data=as_dict)


@controller.route(f'{BASE}/populate-defaults')
def chart_vars_populate_defaults():
  defaults = ktea_client().load_default_values()
  config_man.write_manifest_defaults(defaults)
  return jsonify(data=defaults)


@controller.route(f'{BASE}/<key>')
def manifest_variable_show(key):
  """
  Finds and serializes the chart variable.
  :param key: key used to locate the right chart variable.
  :return: serialized chart variable.
  """
  variable_model = ManifestVariable.find_or_synthesize(key)
  serialized = manifest_vars_serial.full(variable_model)
  return jsonify(data=serialized)


@controller.route(f'{BASE}/commit-apply', methods=['POST'])
def manifest_variables_commit_apply():
  """
  Updates the chart variable with new value.
  :return: status of the update.
  """
  possibly_flat_assignments = jparse()['assignments']

  job_id = job_client.enqueue_action(
    'sdk.action.safely_apply_application_manifest_e2e_action',
    values=utils.flat2deep(possibly_flat_assignments)
  )

  return jsonify(data=dict(job_id=job_id))


@controller.route(f'{BASE}/<key>/validate', methods=['POST'])
def manifest_variable_validate(key):
  """
  Validates the chart variable against
  :param key: key to locate the right chart variable.
  :return: validation status, with tone and message if unsuccessful.
  """
  value = jparse()['value']
  variable_model = ManifestVariable.find_or_synthesize(key)
  eval_result = variable_model.validate(value)
  status = 'valid' if eval_result['met'] else eval_result['tone']
  message = None if eval_result['met'] else eval_result['reason']
  return jsonify(data=dict(status=status, message=message))



from datetime import datetime

from flask import Blueprint, jsonify

from kama_sdk.controllers.ctrl_utils import jparse
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.telem import telem_man
from kama_sdk.serializers import telem_serializers

controller = Blueprint('telem_controller', __name__)

BASE = '/api/telem'


@controller.route(f'{BASE}/prefs')
def prefs_show():
  return jsonify(ping='pong')


@controller.route(f'{BASE}/prefs', methods=['PATCH'])
def prefs_patch():
  return jsonify(ping='pong')


@controller.route(f'{BASE}/config_backups/new', methods=['POST'])
def new_config_backup():
  attrs = jparse()
  config_backup = dict(
    name=attrs.get('name') or 'none',
    event_type='client_triggered',
    data=config_man.serialize(),
    timestamp=str(datetime.now())
  )
  result = telem_man.store_config_backup(config_backup)
  return jsonify(id=str(result.inserted_id), record=config_backup)


@controller.route(f'{BASE}/events')
def list_events():
  return jsonify(data=telem_man.list_events())


@controller.route(f'{BASE}/config_backups/<config_id>')
def get_config_backups(config_id: str):
  if record := telem_man.get_config_backup(config_id):
    serialized = telem_serializers.ser_config_backup_full(record)
    return jsonify(data=serialized)
  else:
    return jsonify(error='dne'), 404


# @controller.route(f'{BASE}/config_backups/<config_id>', methods=['DELETE'])
# def delete_config_backups(config_id: str):
#   if record := telem_man.get_config_backup(config_id):
#     serialized = telem_serializers.ser_config_backup_full(record)
#     return jsonify(data=serialized)
#   else:
#     return jsonify(error='dne'), 404


@controller.route(f'{BASE}/config_backups')
def config_backups_index():
  records = telem_man.list_config_backups()
  serializer = telem_serializers.ser_config_backup
  return jsonify(data=list(map(serializer, records)))



from werkzeug.utils import cached_property

from kama_sdk.core.core.utils import any2bool
from kama_sdk.model.glance.graphic_and_text_glance import GraphicAndTextGlance
from kama_sdk.model.supplier.predicate.predicate import Predicate


class PredicateGlance(GraphicAndTextGlance):
  @cached_property
  def view_type(self) -> str:
    return "resource"

  @cached_property
  def predicate(self):
    return self.inflate_child(
      Predicate,
      prop=PREDICATE_KEY,
      resolve_kod=False
    )

  def eval_result(self) -> bool:
    return any2bool(self.predicate.resolve())

  def content_spec(self):
    success = self.eval_result()
    self.patch({'outcome': success})
    return super(PredicateGlance, self).content_spec()


PREDICATE_KEY = 'predicate'



from datetime import datetime

from kama_sdk.model.operation.operation_state import OperationState


def record(op_state: OperationState):
  return dict(
    status=op_state.status,
    occurred_at=datetime.now(),
    tasks=[
      op_state.preflight_telem,
      *[s.action_telem for s in op_state.step_states]
    ]
  )



from typing import Optional

from kama_sdk.core.core.types import KoD
from kama_sdk.core.core.utils import any2bool
from kama_sdk.model.supplier.base.supplier import Supplier


class Switch(Supplier):
  def resolve(self) -> Optional[KoD]:
    for it in self._compute():
      predicate_kod, value = it.get('break'), it.get('value')
      predicate_outcome = self.resolve_prop_value(predicate_kod)
      if any2bool(predicate_outcome):
        return value
    return None



import time
from typing import List

from werkzeug.utils import cached_property

from kama_sdk.core.core import consts
from kama_sdk.core.core.types import ErrCapture
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError
from kama_sdk.model.action.base.multi_action import MultiAction
from kama_sdk.model.operation.predicate_statuses_computer import PredicateStatusesComputer
from kama_sdk.model.supplier.predicate.predicate import Predicate


class AwaitPredicatesSettledAction(MultiAction):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self._predicates_hack: List[Predicate] = []
    self._sub_actions_hack: List[Action] = []

  def predicates(self) -> List[Predicate]:
    return self.inflate_children(
      Predicate,
      prop=PREDICATES_KEY,
      resolve_kod=False
    )

  @cached_property
  def timeout_seconds(self) -> int:
    return self.get_prop(TIMEOUT_SECONDS_KEY, 180)

  def final_sub_actions(self) -> List[Action]:
    return self._sub_actions_hack

  def perform(self) -> None:
    computer = PredicateStatusesComputer(self.predicates())

    optimists = computer.optimist_predicates
    self._sub_actions_hack = list(map(self.pred2action, optimists))

    start_ts = time.time()
    did_time_out = True

    while time.time() - start_ts < self.timeout_seconds:
      computer.perform_iteration()
      self.update_actions(computer)
      self.raise_for_negative_outcomes(computer)

      if computer.did_finish():
        did_time_out = False
        break
      else:
        time.sleep(2)

    raise_for_timeout(did_time_out)

  def find_action_by_pred(self, predicate_id: str) -> Action:
    finder = lambda action: action.config['predicate_id'] == predicate_id
    return next(filter(finder, self.final_sub_actions()))

  def update_actions(self, computer: PredicateStatusesComputer):
    for predicate in computer.optimist_predicates:
      evaluation = computer.find_eval(predicate.id())
      action = self.find_action_by_pred(predicate.id())
      action.set_status(consts.pos if evaluation['met'] else consts.rng)

  def pred2action(self, predicate: Predicate) -> Action:
    action = Action(dict(
      id=predicate.id(),
      title=predicate.title,
      info=predicate.info,
      predicate_id=predicate.id()
    ))
    action.parent = self
    return action

  @staticmethod
  def raise_for_negative_outcomes(computer: PredicateStatusesComputer):
    if computer.did_fail():
      predicate = computer.culprit_predicate()
      raise FatalActionError(ErrCapture(
        type='predicate_settle_negative',
        reason=predicate.reason(),
        extras=dict(
          predicate_id=predicate.id(),
          predicate_kind=predicate.kind(),
          **predicate.error_extras()
        )
      ))


def raise_for_timeout(timed_out: bool):
  if timed_out:
    raise FatalActionError(ErrCapture(
      type='predicate_settle_timeout',
      reason='Time'
    ))


PREDICATES_KEY = 'predicates'
TIMEOUT_SECONDS_KEY = 'timeout_seconds'



http_api_ktea = 'server'
local_exec_ktea = 'local_executable'
virtual_ktea = 'virtual_ktea'

FILE_PREFIX = "file"
SUPPLIER_PREFIX = "get"

pos = 'positive'
neg = 'negative'
rng = 'running'
brk = 'broken'
idl = 'idle'

TARGET_STANDARD = 'standard'
TARGET_INLIN = 'inline'
TARGET_STATE = 'state'
TARGET_PREFS = 'prefs'

TARGET_TYPES = [TARGET_STANDARD, TARGET_INLIN, TARGET_STATE, TARGET_PREFS]
DEFAULT_TARGET = TARGET_STANDARD

MAIN_WORKER = 'default'
TELEM_WORKER = 'telem'


statuses = [pos, neg, rng, idl]



def default_model_classes():
  from kama_sdk.model.supplier.base.misc_suppliers import FormattedDateSupplier
  from kama_sdk.model.variable.manifest_variable import ManifestVariable
  from kama_sdk.model.input.generic_input import GenericInput
  from kama_sdk.model.supplier.ext.misc.latest_vendor_injection_supplier import LatestVendorInjectionSupplier
  from kama_sdk.model.input.slider_input import SliderInput
  from kama_sdk.model.operation.operation import Operation
  from kama_sdk.model.operation.stage import Stage
  from kama_sdk.model.operation.step import Step
  from kama_sdk.model.operation.field import Field
  from kama_sdk.model.variable.generic_variable import GenericVariable
  from kama_sdk.model.supplier.ext.biz.resource_selector import ResourceSelector
  from kama_sdk.model.operation.operation_run_simulator import OperationRunSimulator
  from kama_sdk.model.action.base.multi_action import MultiAction
  from kama_sdk.model.input.checkboxes_input import CheckboxesInput
  from kama_sdk.model.input.checkboxes_input import CheckboxInput
  from kama_sdk.model.supplier.predicate.multi_predicate import MultiPredicate
  from kama_sdk.model.supplier.base.supplier import Supplier
  from kama_sdk.model.supplier.ext.misc.http_data_supplier import HttpDataSupplier
  from kama_sdk.model.supplier.ext.biz.resources_supplier import ResourcesSupplier
  from kama_sdk.model.input.checkboxes_input import SelectInput
  from kama_sdk.model.supplier.ext.misc.service_endpoint_supplier import ServiceEndpointSupplier
  from kama_sdk.model.supplier.ext.misc.random_string_supplier import RandomStringSupplier
  from kama_sdk.model.supplier.ext.biz.master_config_supplier import MasterConfigSupplier
  from kama_sdk.model.action.base.action import Action
  from kama_sdk.model.supplier.ext.prometheus.prom_vector_to_first_value_supplier import PromVectorToFirstValueSupplier
  from kama_sdk.model.supplier.ext.prometheus.prom_matrix_to_series_supplier import PromMatrixToSeriesSupplier
  from kama_sdk.model.humanizer.quantity_humanizer import QuantityHumanizer
  from kama_sdk.model.humanizer.cores_humanizer import CoresHumanizer
  from kama_sdk.model.humanizer.bytes_humanizer import BytesHumanizer
  from kama_sdk.model.action.ext.manifest.await_outkomes_settled_action import AwaitOutkomesSettledAction
  from kama_sdk.model.action.ext.manifest.await_predicates_settled_action import AwaitPredicatesSettledAction
  from kama_sdk.model.action.ext.manifest.kubectl_apply_action import KubectlApplyAction
  from kama_sdk.model.action.ext.manifest.template_manifest_action import TemplateManifestAction
  from kama_sdk.model.action.ext.update.update_actions import FetchUpdateAction
  from kama_sdk.model.action.ext.update.update_actions import CommitKteaFromUpdateAction
  from kama_sdk.model.action.ext.misc.run_predicates_action import RunPredicateAction
  from kama_sdk.model.action.ext.misc.run_predicates_action import RunPredicatesAction

  from kama_sdk.model.action.ext.misc.wait_action import WaitAction
  from kama_sdk.model.action.ext.misc.delete_resources_action import DeleteResourceAction
  from kama_sdk.model.action.ext.misc.delete_resources_action import DeleteResourcesAction
  from kama_sdk.model.action.ext.manifest.kubectl_dry_run_action import KubectlDryRunAction
  from kama_sdk.model.action.ext.misc.create_backup_action import CreateBackupAction
  from kama_sdk.model.supplier.base.props_supplier import PropsSupplier
  from kama_sdk.model.supplier.base.switch import Switch
  from kama_sdk.model.action.ext.update.fetch_latest_injection_action import FetchLatestInjectionsAction
  from kama_sdk.model.supplier.predicate.format_predicate import FormatPredicate
  from kama_sdk.model.supplier.predicate.common_predicates import TruePredicate
  from kama_sdk.model.supplier.predicate.common_predicates import FalsePredicate
  from kama_sdk.model.supplier.predicate.predicate import Predicate
  from kama_sdk.model.action.ext.manifest.patch_manifest_vars_action import PatchManifestVarsAction
  from kama_sdk.model.action.ext.update.update_actions import LoadVarDefaultsFromKtea
  from kama_sdk.model.action.ext.manifest.patch_manifest_vars_action import WriteManifestVarsAction
  from kama_sdk.model.supplier.base.misc_suppliers import SumSupplier
  from kama_sdk.model.supplier.base.misc_suppliers import MergeSupplier
  from kama_sdk.model.supplier.base.misc_suppliers import ListFlattener
  from kama_sdk.model.supplier.base.misc_suppliers import ListPluck
  from kama_sdk.model.supplier.base.misc_suppliers import IfThenElse
  from kama_sdk.model.supplier.ext.prometheus.prom_supplier import PromDataSupplier
  from kama_sdk.model.supplier.ext.misc.redactor import Redactor
  from kama_sdk.model.supplier.ext.vis.series_summary_supplier import SeriesSummarySupplier
  from kama_sdk.model.supplier.ext.vis.pod_statuses_supplier import PodStatusSummariesSupplier
  from kama_sdk.model.supplier.base.provider import Provider
  from kama_sdk.model.supplier.base.misc_suppliers import ListFilter
  from kama_sdk.model.concern.concern import Concern
  from kama_sdk.model.concern.concern_table import ConcernTable

  from kama_sdk.model.concern.concern_set import ConcernSet
  from kama_sdk.model.concern.concern_view_adapter import ConcernViewAdapter
  from kama_sdk.model.concern.concern_card_adapter import ConcernCardAdapter
  from kama_sdk.model.concern.concern_detail_adapter import ConcernDetailAdapter
  from kama_sdk.model.concern.concern_super_set import ConcernSuperSet
  from kama_sdk.model.concern.concern_grid import ConcernGrid
  from kama_sdk.model.concern.concern_attr_adapter import ConcernAttrAdapter

  return [
    Operation,
    Stage,
    Step,
    Field,

    GenericVariable,
    ManifestVariable,
    ResourceSelector,

    GenericInput,
    SliderInput,
    SelectInput,
    CheckboxesInput,
    CheckboxInput,

    Predicate,
    FormatPredicate,
    MultiPredicate,
    TruePredicate,
    FalsePredicate,

    Supplier,
    Provider,
    PropsSupplier,
    FormattedDateSupplier,
    Switch,
    MergeSupplier,
    HttpDataSupplier,
    ResourcesSupplier,
    RandomStringSupplier,
    MasterConfigSupplier,
    SumSupplier,
    ServiceEndpointSupplier,
    LatestVendorInjectionSupplier,
    ListFlattener,
    ListFilter,
    ListPluck,
    IfThenElse,
    Redactor,

    PromDataSupplier,
    PromMatrixToSeriesSupplier,
    PromVectorToFirstValueSupplier,
    SeriesSummarySupplier,
    PodStatusSummariesSupplier,

    Action,
    MultiAction,
    PatchManifestVarsAction,
    WriteManifestVarsAction,
    LoadVarDefaultsFromKtea,
    FetchLatestInjectionsAction,
    AwaitOutkomesSettledAction,
    AwaitPredicatesSettledAction,
    KubectlApplyAction,
    TemplateManifestAction,
    FetchUpdateAction,
    CommitKteaFromUpdateAction,
    RunPredicateAction,
    RunPredicatesAction,
    WaitAction,
    DeleteResourceAction,
    DeleteResourcesAction,
    KubectlDryRunAction,
    CreateBackupAction,

    Concern,
    ConcernSuperSet,
    ConcernSet,
    ConcernTable,
    ConcernGrid,
    ConcernViewAdapter,
    ConcernCardAdapter,
    ConcernDetailAdapter,
    ConcernAttrAdapter,

    QuantityHumanizer,
    BytesHumanizer,
    CoresHumanizer,

    OperationRunSimulator
  ]



from flask import Blueprint, jsonify

from kama_sdk.core.core import job_client, updates_man
from kama_sdk.core.telem import telem_man

controller = Blueprint('updates_controller', __name__)

BASE_PATH = '/api/updates'


@controller.route(BASE_PATH)
def get_all_updates():
  telem_man.upload_status()
  updates = updates_man.fetch_all()
  return jsonify(data=updates)


@controller.route(f'{BASE_PATH}/next-available')
def fetch_next_available():
  update_or_none = updates_man.next_available()
  return jsonify(data=update_or_none)


@controller.route(f'{BASE_PATH}/<update_id>')
def show_update(update_id):
  update = updates_man.fetch_update(update_id)
  return jsonify(data=update)


@controller.route(f'{BASE_PATH}/<update_id>/preview')
def preview_update(update_id):
  update = updates_man.fetch_update(update_id)
  preview_bundle = updates_man.preview(update)
  return jsonify(preview_bundle)


@controller.route(f'{BASE_PATH}/<update_id>/apply', methods=['POST'])
def install_update(update_id):
  job_id = job_client.enqueue_action(
    'sdk.action.apply_update_e2e_action',
    update_id= update_id
  )
  return jsonify(data=(dict(job_id=job_id)))


@controller.route(f'{BASE_PATH}/outcomes')
def list_past_updates():
  outcomes = telem_man.list_events()
  return jsonify(data=outcomes)



from kama_sdk.model.input.generic_input import GenericInput


def in_variable(ginput: GenericInput):
  return dict(
    type=ginput.kind(),
    options=ginput.serialize_options(),
    extras=ginput.extras()
  )



from functools import cached_property
from typing import Optional, List, Dict

from kama_sdk.core.core.types import PromSeriesDataPoint, SimpleSeriesSummary
from kama_sdk.model.humanizer.quantity_humanizer import QuantityHumanizer
from kama_sdk.model.supplier.base.supplier import Supplier


class SeriesSummarySupplier(Supplier):

  @cached_property
  def time_series(self) -> List[PromSeriesDataPoint]:
    v = self.source_data()
    # print("SOURCE DATA")
    # print(v)
    return v

  @cached_property
  def reducer_type(self) -> str:
    return self.get_prop(REDUCER_FUNC_KEY, 'last')

  def change_direction(self) -> str:
    try:
      count = len(self.time_series)
      if count >= 2:
        last = _last_value(self.time_series)
        avg = _series_avg(self.time_series)
        return 'up' if last - avg > 0 else 'down'
    except:
      return ''

  @cached_property
  def good_direction(self) -> str:
    return self.get_prop(GOOD_DIRECTION_KEY, 'up')

  @cached_property
  def humanizer(self) -> QuantityHumanizer:
    return self.inflate_child(
      QuantityHumanizer,
      prop=HUMANIZER_KEY,
      safely=True
    ) or QuantityHumanizer({})

  # noinspection PyBroadException
  def summary_quant(self) -> Optional[float]:
    try:
      if self.reducer_type == 'last':
        return _last_value(self.time_series)
      elif self.reducer_type == 'sum':
        return _series_sum(self.time_series)
    except:
      return None

  def summary_value(self) -> str:
    return self.humanizer.humanize_expr(self.summary_quant())

  def humanized_series(self) -> List[Dict]:
    return [humanize_datapoint(d, self.humanizer) for d in self.time_series]

  def resolve(self) -> Optional[SimpleSeriesSummary]:
    if self.time_series:
      return SimpleSeriesSummary(
        series=self.time_series,
        humanized_series=self.humanized_series(),
        direction=self.change_direction(),
        good_direction=self.good_direction,
        summary_value=self.summary_value()
      )
    else:
      return None


def humanize_datapoint(datapoint, humanizer: QuantityHumanizer):
  return {
    **datapoint,
    'value': humanizer.humanize_quantity(datapoint['value'])
  }


def _last_value(series: List[Dict]) -> Optional[float]:
  if len(series) > 0:
    return series[len(series) - 1]['value']
  else:
    return None


def _series_sum(series: List[PromSeriesDataPoint]) -> float:
  return sum([float(d['value'] or 0) for d in series])


def _series_avg(series: List[PromSeriesDataPoint]) -> float:
  try:
    return _series_sum(series) / len(series)
  except ValueError:
    return 0.0


DATA_KEY = 'time_series_data'
REDUCER_FUNC_KEY = 'reducer'
HUMANIZER_KEY = 'humanizer'
GOOD_DIRECTION_KEY = 'good_direction'



from typing import Dict

from kama_sdk.model.action.ext.misc.run_predicates_action import RunPredicatesAction


def serialize_std(check: RunPredicatesAction) -> Dict:
  return dict(
    id=check.id(),
    title=check.title,
    info=check.info,
    length=len(check.predicates()),
    tags=check.tags
  )



import time

from werkzeug.utils import cached_property

from kama_sdk.model.action.base.action import Action


class WaitAction(Action):

  @cached_property
  def title(self) -> str:
    return f"Wait {self.duration_seconds} seconds"

  @cached_property
  def info(self) -> str:
    return f"Holds the thread for the requested duration"

  def duration_seconds(self) -> int:
    return int(self.get_prop('duration_seconds', 3))

  def perform(self) -> None:
    time.sleep(self.duration_seconds())



from typing import Any

import requests
from werkzeug.utils import cached_property

from kama_sdk.model.supplier.base.supplier import Supplier


class HttpDataSupplier(Supplier):

  ENDPOINT_KEY = 'endpoint'

  @cached_property
  def endpoint(self):
    return self.resolve_prop(self.ENDPOINT_KEY, missing='warn')

  @cached_property
  def output_format(self):
    super_value = super(HttpDataSupplier, self).output_format
    return super_value or 'status_code'

  # noinspection PyBroadException
  def _compute(self) -> Any:
    try:
      response = requests.get(self.endpoint)
      body_dict = {}
      try:
        body_dict = response.json()
      except:
        pass
      return dict(
        status_code=response.status_code,
        body=body_dict
      )
    except:
      return None



from typing import List, Optional

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import PredEval
from kama_sdk.core.core.utils import any2bool
from kama_sdk.model.supplier.predicate.predicate import Predicate

class PredicateStatusesComputer:
  def __init__(self, predicates: List[Predicate]):
    self.predicates: List[Predicate] = predicates
    self.evaluations: List[PredEval] = []
    self._did_succeed: Optional[bool] = None

  def perform_iteration(self):
    if len(self.predicates) > 0:
      self.evaluate_batch(self.optimist_predicates)
      if all_conditions_met(self.optimist_evaluations()):
        self.conclude(True)

      self.evaluate_batch(self.pessimist_predicates)
      if any_condition_met(self.pessimist_evaluations()):
        self.conclude(False)
    else:
      self.conclude(True)

  def evaluate_batch(self, predicates: List[Predicate]):
    for predicate in predicates:
      previous_eval = self.find_eval(predicate.id())
      if needs_recomputing(previous_eval, predicate.is_optimist()):
        new_eval = eval_pred(predicate)
        self.process_new_evaluation(new_eval)
        if can_halt_early(new_eval, predicate):
          return

  def find_predicate(self, predicate_id: str) -> Predicate:
    finder = lambda predicate: predicate.id() == predicate_id
    return next(filter(finder, self.predicates), None)

  def find_eval(self, predicate_id: str) -> Optional[PredEval]:
    finder = lambda pred_eval: pred_eval['predicate_id'] == predicate_id
    return next(filter(finder, self.evaluations), None)

  def process_new_evaluation(self, new_eval: PredEval):
    prev_eval = self.find_eval(new_eval['predicate_id'])
    if prev_eval:
      prev_eval['met'] = new_eval['met']
      prev_eval['reason'] = new_eval['reason']
    else:
      self.evaluations.append(new_eval)

  @cached_property
  def optimist_predicates(self) -> List[Predicate]:
    return [p for p in self.predicates if p.is_optimist()]

  @cached_property
  def pessimist_predicates(self) -> List[Predicate]:
    return [p for p in self.predicates if not p.is_optimist()]

  def optimist_evaluations(self) -> List[PredEval]:
    decider = lambda e: self.find_predicate(e['predicate_id']).is_optimist()
    return list(filter(decider, self.evaluations))

  def pessimist_evaluations(self) -> List[PredEval]:
    decider = lambda e: not self.find_predicate(e['predicate_id']).is_optimist()
    return list(filter(decider, self.evaluations))

  def conclude(self, success: bool):
    self._did_succeed = success

  def did_finish(self) -> bool:
    return self._did_succeed is not None

  def did_fail(self):
    return self.did_finish() and not self._did_succeed

  def culprit_predicate(self) -> Predicate:
    filterer = filter(lambda e: e['met'], self.pessimist_evaluations())
    return self.find_predicate(next(filterer)['predicate_id'])


def eval_pred(predicate: Predicate) -> PredEval:
  eval_result = any2bool(predicate.resolve())
  return PredEval(
    predicate_id=predicate.id(),
    met=eval_result,
    name=predicate.title,
    reason=predicate.reason()
  )


def can_halt_early(pred_eval: PredEval, predicate: Predicate) -> bool:
  return pred_eval['met'] and not predicate.is_optimist()


def needs_recomputing(pred_eval: Optional[PredEval], is_optimist) -> bool:
  return not pred_eval or not is_optimist or not pred_eval['met']


def all_conditions_met(evaluations: List[PredEval]) -> bool:
  return set([s['met'] for s in evaluations]) == {True}


def any_condition_met(evaluations: List[PredEval]) -> bool:
  return True in [s['met'] for s in evaluations]



from k8kat.utils.main import units

from kama_sdk.model.humanizer.quantity_humanizer import QuantityHumanizer


class BytesHumanizer(QuantityHumanizer):

  @staticmethod
  def _humanize_quantity(value: float) -> float:
    bytes_in_unit, _ = units.mem_quant_mult(value or 0)
    return bytes_in_unit

  def _humanize_expr(self, raw_value: float) -> str:
    _, unit = units.mem_quant_mult(raw_value or 0)
    value_str = self.humanize_quantity(raw_value)
    return f"{value_str}{unit}b"



from typing import Dict, Optional, Union, List

from kama_sdk.core.telem.telem_perms import TelemPerms


def if_allowed(perms: TelemPerms, prop_name: str, value) -> Optional:
  can_share = perms.can_share_prop(prop_name)
  if can_share:
    return value() if callable(value) else value
  return None


def redact_for_perms(perms, prefix, dirty_dict: Union[Dict, List]) -> Dict:
  new_dict = {}
  for key, value in dirty_dict.items():
    full_key, sub_key = f'{prefix}{key}', f'{prefix}{key}.'
    if type(value) == list:
      child_sanitizer = lambda itm: redact_for_perms(perms, sub_key, itm)
      comp_children = lambda: list(map(child_sanitizer, value))
      new_dict[key] = if_allowed(perms, full_key, comp_children)
    elif type(value) == dict:
      comp = lambda: redact_for_perms(perms, sub_key, value)
      new_dict[key] = if_allowed(perms, full_key, comp)
    else:
      new_dict[key] = if_allowed(perms, full_key, value)
  return new_dict


def redact_op_outcome(op_outcome: Dict) -> Dict:
  perms = TelemPerms()
  return redact_for_perms(perms, 'operation_outcome.', op_outcome)



import json
import traceback
from typing import Optional, Dict, Callable

from rq import Queue
from rq.exceptions import NoSuchJobError
from rq.job import Job
from werkzeug.utils import cached_property

from kama_sdk.core.core import consts
from kama_sdk.core.core.types import ActionStatusDict, KoD
from kama_sdk.worker import conn

class JobProgressWrapper:
  def __init__(self, job: Job):
    self.job = job

  @cached_property
  def meta(self) -> Dict:
    return self.job.meta or {}

  @cached_property
  def progress_bundle(self) -> Optional[ActionStatusDict]:
    serialized = self.meta.get('progress')
    return json.loads(serialized) if serialized else None

  @cached_property
  def result(self) -> Dict:
    return self.job.result

  def status(self):
    if self.job.is_finished or self.job.is_failed:
      backup_status = consts.pos if self.job.is_finished else consts.neg
      action_progress = self.progress_bundle
      if action_progress:
        expl_status = action_progress.get('status')
        if expl_status not in [consts.pos, consts.neg]:
          print(f"[job_client] danger done bad action status {expl_status}")
        return expl_status
      return backup_status
    else:
      backup_status = 'running'
      action_progress = self.progress_bundle
      if action_progress:
        expl_status = action_progress.get('status')
        if not expl_status == consts.rng:
          print(f"[job_client] danger running bad action status {expl_status}")
          return expl_status
      return backup_status


main_queue = Queue(consts.MAIN_WORKER, connection=conn)
telem_queue = Queue(consts.TELEM_WORKER, connection=conn)


def enqueue_action(kod: KoD, **kwargs) -> str:
  return enqueue_func(load_and_perform_action, kod, **kwargs)


def enqueue_func(func: Callable, *args, **kwargs) -> str:
  job = main_queue.enqueue(func, *args, **kwargs)
  return job.get_id()


def enqueue_or_run_telem_func(enq: bool, func: Callable, *args, **kwargs) -> str:
  if enq:
    job = telem_queue.enqueue(func, *args, **kwargs)
    return job.get_id()
  else:
    return func(*args, **kwargs)


def find_job(job_id: str) -> Optional[Job]:
  try:
    return Job.fetch(job_id, connection=conn)
  except NoSuchJobError:
    return None


def job_status(job_id: str) -> JobProgressWrapper:
  job = find_job(job_id)
  return JobProgressWrapper(job)


def load_and_perform_action(kod: KoD, **kwargs):
  from kama_sdk.model.action.base.action import Action
  action: Action = Action.inflate(kod, patches=kwargs)
  # print("----------- inflated acction")
  # print(action.config)
  try:
    action.run()
    print(f"[job_client] action={action.id()} succeeded")
  except Exception:
    err = traceback.format_exc()
    print(f"[job_client] action={action.id()} failed: \n\n {err}")
    return None



from typing import Optional, Dict

from werkzeug.utils import cached_property

from kama_sdk.core.core import updates_man, hub_api_client
from kama_sdk.core.core.types import UpdateDict, KteaDict, ErrCapture
from kama_sdk.core.ktea.ktea_provider import ktea_client
from kama_sdk.model.action.base.action import Action
from kama_sdk.model.action.base.action_errors import FatalActionError


class LoadVarDefaultsFromKtea(Action):

  def ktea(self) -> KteaDict:
    return self.resolve_prop('ktea')

  def perform(self, **kwargs) -> Optional[Dict]:
    client = ktea_client(ktea=self.ktea())
    variable_defaults = client.load_default_values()
    raise_if_none_defaults(variable_defaults)
    return dict(variable_defaults=variable_defaults)


def raise_if_none_defaults(defaults: Optional[Dict]):
  if not defaults:
    raise FatalActionError(ErrCapture(
      type='load_defaults',
      name='empty_resp',
      reason='ktea returned nil for defaults'
    ))


class FetchUpdateAction(Action):

  def update_id(self) -> str:
    return self.resolve_prop('update_id')

  def perform(self, **kwargs) -> Optional[UpdateDict]:
    update = updates_man.fetch_update(self.update_id())
    self.raise_if_none(update)
    return dict(update=update)

  def raise_if_none(self, update: Optional[UpdateDict]) -> None:
    if not update:
      host = hub_api_client.backend_host()
      update_id = self.update_id()
      raise FatalActionError(ErrCapture(
        type='fetch_update',
        name='could_not_fetch',
        reason=f"fetch failed update id={update_id} host {host}",
        extras=dict(host=host, update_id=update_id)
      ))


class CommitKteaFromUpdateAction(Action):

  @cached_property
  def update(self) -> UpdateDict:
    return self.get_prop('update')

  def perform(self):
    updates_man.commit_new_ktea(self.update)



from typing import Dict, Any, Optional, List, TypeVar

from werkzeug.utils import cached_property

from kama_sdk.core.core import utils
from kama_sdk.model.supplier.base import supplier
from kama_sdk.model.supplier.base.supplier import Supplier

T = TypeVar('T', bound='Predicate')

class Predicate(Supplier):

  @classmethod
  def inflate_with_literal(cls, expr: str, **kwargs) -> Optional[T]:
    operator, rest = expr.split("?")
    if '|' in rest:
      parts = rest.split('|')
      other_params = {
        CHECK_AGAINST_KEY: parts[1],
        CHALLENGE_KEY: parts[0]
      }
    else:
      other_params = {CHECK_AGAINST_KEY: rest}

    return cls.inflate_with_config({
      OPERATOR_KEY: operator,
      **other_params
    }, **kwargs)

  @cached_property
  def serializer_type(self) -> str:
    return supplier.SER_IDENTITY

  def many_policy(self) -> str:
    return self.resolve_prop(ON_MANY_KEY, lookback=0)

  def challenge(self) -> Any:
    # we need a new @wiz_property with custom caching logic
    return self.resolve_prop(CHALLENGE_KEY)

  def check_against(self) -> Optional[Any]:
    return self.resolve_prop(CHECK_AGAINST_KEY)

  def operator(self) -> str:
    return self.get_prop(OPERATOR_KEY, '==')

  def is_optimist(self) -> bool:
    return self.get_prop(IS_OPTIMISTIC_KEY, False)

  def is_pessimist(self) -> bool:
    return not self.is_optimist()

  def tone(self) -> str:
    return self.get_prop(TONE_KEY, 'error')

  def reason(self) -> str:
    return self.get_prop(REASON_KEY, '')

  def resolve(self) -> bool:
    return perform_comparison(
      self.operator(),
      self.challenge(),
      self.check_against(),
      self.many_policy()
    )

  def error_extras(self) -> Dict:
    return self.resolve_prop(ERROR_EXTRAS_KEY, depth=100) or {}


def perform_comparison(_name: str, challenge, against, on_many) -> bool:
  if utils.listlike(challenge) and on_many:
    run = lambda v: standard_comparison(_name, v, against)
    results: List[bool] = list(map(run, challenge))
    if on_many == 'each_true':
      return set(results) == {True}
    elif on_many == 'each_false':
      return set(results) == {False}
    elif on_many == 'some_true':
      return True in results
    elif on_many == 'some_false':
      return False in results
    else:
      print(f"[kama_sdk:predicate] bad many policy {on_many}")
  else:
    return standard_comparison(_name, challenge, against)


def standard_comparison(_name: str, challenge: Any, against: Any) -> bool:
  challenge = utils.unmuck_primitives(challenge)
  against = utils.unmuck_primitives(against)

  try:
    if _name in ['equals', 'equal', 'eq', '==', '=']:
      return challenge == against

    elif _name in ['not-equals', 'not-equal', 'neq', '!=', '=/=']:
      return not challenge == against

    elif _name in ['is-in', 'in']:
      return challenge in against

    elif _name in ['contains']:
      return against in challenge

    elif _name in ['only', 'contains-only']:
      return set(challenge) == {against}

    elif _name in ['is-greater-than', 'greater-than', 'gt', '>']:
      return challenge > against

    elif _name in ['gte', '>=']:
      return challenge >= against

    elif _name in ['is-less-than', 'less-than', 'lt', '<']:
      return challenge < against

    elif _name in ['lte', '<=']:
      return challenge <= against

    if _name in ['truthy', 'truthiness']:
      return utils.any2bool(challenge)

    elif _name in ['presence', 'defined', 'is-defined']:
      return bool(challenge)

    elif _name in ['undefined', 'is-undefined']:
      return not challenge

    print(f"Don't know operator {_name}")
    return False
  except:
    return False


CHALLENGE_KEY = 'challenge'
OPERATOR_KEY = 'operator'
CHECK_AGAINST_KEY = 'check_against'
ON_MANY_KEY = 'many_policy'
TONE_KEY = 'tone'
REASON_KEY = 'reason'
IS_OPTIMISTIC_KEY = 'optimistic'
TAGS = 'tags'
ERROR_EXTRAS_KEY = 'error_extras'
APP_STATUS_PRED_ID = 'sdk.predicate.app_status'




from typing import Any, Optional

from werkzeug.utils import cached_property

from kama_sdk.model.supplier.base.supplier import Supplier


class PropsSupplier(Supplier):

  def orig_output_format(self) -> Optional[str]:
    return self.config.get('output')

  @cached_property
  def default_lookback(self):
    return self.config.get('lookback', True)

  def markers(self):
    orig = self.orig_output_format()
    if orig and type(orig) == str:
      if jq_accessor_token in orig:
        return orig.split(jq_accessor_token)[1]

  def source_data(self) -> Optional[Any]:
    output_fmt = self.orig_output_format()
    if jq_accessor_token in output_fmt:
      master_key = output_fmt.split(jq_accessor_token)[0]
    elif legacy_accessor_token in output_fmt:
      master_key = output_fmt.split(legacy_accessor_token)[0]
    else:
      master_key = output_fmt
    return self.resolve_prop(master_key)

  @cached_property
  def output_format(self):
    orig = self.orig_output_format()
    if orig and type(orig) == str:
      if jq_accessor_token in orig:
        return orig.split(jq_accessor_token)[1]
      elif legacy_accessor_token in orig:
        return orig.split(legacy_accessor_token)[1]
      else:
        return None
    else:
      return orig

  @cached_property
  def serializer_type(self) -> str:
    if legacy_accessor_token in self.orig_output_format():
      return 'legacy'
    else:
      return 'jq'


jq_accessor_token = '->'
legacy_accessor_token = '=>'



from typing import Dict

from kama_sdk.model.error.diagnosis_actionable import DiagnosisActionable
from kama_sdk.model.error.error_diagnosis import ErrorDiagnosis


def ser_err_diagnosis(diagnosis: ErrorDiagnosis) -> Dict:
  actionables = list(map(ser_embedded_actionable, diagnosis.actionables()))

  return dict(
    id=diagnosis.id(),
    title=diagnosis.title,
    info=diagnosis.info,
    actionables=actionables,
  )


def ser_embedded_actionable(actionable: DiagnosisActionable) -> Dict:
  return dict(
    id=actionable.id(),
    title=actionable.title,
    info=actionable.info,
    operation_id=actionable.operation_id,

  )



from typing import Dict

from flask import Blueprint, jsonify, request

from kama_sdk.controllers.ctrl_utils import jparse
from kama_sdk.core.core import job_client, utils, consts
from kama_sdk.core.telem import telem_man
from kama_sdk.model.operation.operation import Operation
from kama_sdk.model.operation.operation_state import OperationState, operation_states
from kama_sdk.model.operation.stage import Stage
from kama_sdk.model.operation.step import Step
from kama_sdk.serializers import operation_serial as operation_serial, step_serial

OPERATIONS_PATH = '/api/operations'
OPERATION_PATH = f'/{OPERATIONS_PATH}/<operation_id>'

STAGES_PATH = f'{OPERATION_PATH}/stages'
STAGE_PATH = f'{STAGES_PATH}/<stage_id>'

STEPS_PATH = f'{STAGE_PATH}/steps'
STEP_PATH = f'{STEPS_PATH}/<step_id>'

FIELDS_PATH = f'{STEP_PATH}/fields'
FIELD_PATH = f'{FIELDS_PATH}/<field_id>'

controller = Blueprint('operations_controller', __name__)


@controller.route(OPERATIONS_PATH)
def operations_index():
  """
  Lists all existing Operations for a local app, except system ones.
  :return: list of minimally serialized Operation objects.
  """
  all_ops = Operation.inflate_all()
  operations_list = [op for op in all_ops if not op.is_system()]
  dicts = [operation_serial.ser_standard(c) for c in operations_list]
  return jsonify(data=dicts)


@controller.route(OPERATION_PATH)
def operations_show(operation_id):
  """
  Finds and returns a particular Operation using operation_id.
  :param operation_id: operation id to search by.
  :return: fully serialized Operation object.
  """
  operation = find_operation(operation_id)
  return jsonify(data=operation_serial.ser_full(operation))


@controller.route(f'{OPERATIONS_PATH}/osts')
def operations_ost_index():
  """
  Generates a list of currently available OperationStates.
  :return: list of currently available OperationStates.
  """
  return jsonify(data=operation_states)


@controller.route(f'{OPERATION_PATH}/generate-ost', methods=['POST'])
def operations_gen_ost(operation_id):
  """
  Generates a new OST (random 10 character string).
  :return: new OST.
  """
  uuid = OperationState.gen(operation_id)
  return jsonify(data=uuid)


@controller.route(f"{OPERATION_PATH}/eval-preflight", methods=['POST'])
def eval_preflight(operation_id):
  """
  Finds the Prerequisite with a matching operation_id and prerequisite_id,
  and evaluates it.
  :param operation_id: operation id to search by.
  :return: dict containing results of evaluation.
  """
  operation = find_operation(operation_id)
  action_kod = operation.preflight_action_kod
  if action_kod:
    job_id = job_client.enqueue_action(action_kod)
    return jsonify(data=dict(job_id=job_id))
  else:
    return jsonify(data=dict(job_id=None))


@controller.route(f"{STEP_PATH}/refresh", methods=['POST'])
def step_refresh(operation_id, stage_id, step_id):
  """
  Finds the Step with a matching operation_id, stage_id and step_id.
  :param operation_id: operation id to search by.
  :param stage_id: stage id to search by.
  :param step_id: step id to search by.
  :return: serialized Step object.
  """
  values: Dict = utils.flat2deep(jparse()['values'])
  # print(f"HEYE GOING IN WITH VALUES \n\n{values}")
  op_state = find_op_state(operation_id=operation_id)
  step = find_step(operation_id, stage_id, step_id)
  find_op_state().gen_step_state(step, keep=False)
  asgs = step.partition_vars(values, op_state)
  serialized = step_serial.ser_refreshed(step, values, op_state)
  return jsonify(data=dict(
    step=serialized,
    manifest_assignments=asgs.get(consts.TARGET_STANDARD)
  ))


@controller.route(f"{STEP_PATH}/preview-chart-assignments", methods=['POST'])
def step_preview_chart_assigns(operation_id, stage_id, step_id):
  """
  Returns the chart assignments that would be committed if the step were submitted
  with current user input.
  :param operation_id: operation id used to locate the right step
  :param stage_id: stage id used to locate the right step
  :param step_id: step id used to locate the right step
  :return: dictionary with chart assigns.
  """
  values = utils.flat2deep(jparse()['values'])
  op_state = find_op_state(operation_id=operation_id)
  step = find_step(operation_id, stage_id, step_id)
  op_state.gen_step_state(step, keep=False)
  asgs = step.partition_vars(values, op_state)
  return jsonify(data=asgs[consts.TARGET_STANDARD])


@controller.route(f"{STEP_PATH}/run", methods=['POST'])
def step_run(operation_id, stage_id, step_id):
  """
  Submits a step. This includes:
    1. Finding the appropriate OperationState
    2. Appending a new StepState to OperationState
    3. Committing the step (See docs for step commit)
    4. Updating the StepState with the details of the commit outcome
  :param operation_id: operation id to search by.
  :param stage_id: stage id to search by.
  :param step_id: step id to search by.
  :return: dict containing submit status, message and logs.
  """
  values = utils.flat2deep(jparse()['values'])
  step = find_step(operation_id, stage_id, step_id)
  op_state = find_op_state(operation_id=operation_id)
  step_state = op_state.gen_step_state(step)
  job_id = step.run(values, step_state)
  return jsonify(data=dict(job_id=job_id))


@controller.route(f'{STEP_PATH}/next')
def steps_next_id(operation_id, stage_id, step_id):
  """
  computes and returns the id of the next step.
  :param operation_id: operation id to locate the right step.
  :param stage_id: stage id to locate the right step.
  :param step_id: step id to locate the right step.
  :return: computed id of next step or "done" if no more steps left.
  """
  stage = find_stage(operation_id, stage_id)
  step = find_step(operation_id, stage_id, step_id)
  op_state = find_op_state()
  result = stage.next_step_id(step, op_state)
  return jsonify(step_id=result)


@controller.route(f'{FIELD_PATH}/validate', methods=['POST'])
def step_field_validate(operation_id, stage_id, step_id, field_id):
  value = jparse()['value']
  step = find_step(operation_id, stage_id, step_id)
  op_state = find_op_state(operation_id=operation_id)
  eval_result = step.validate_field(field_id, value, op_state)
  status = 'valid' if eval_result['met'] else eval_result['tone']
  message = None if eval_result['met'] else eval_result['reason']
  return jsonify(data=dict(status=status, message=message))


@controller.route(f'{OPERATIONS_PATH}/mark-finished', methods=['POST'])
def operation_mark_finished():
  """
  Syncs telemetry information with the database if permissions allow to do so.
  Deletes OperationState after its done.
  :return: success or failure status depending if managed to find and delete
  the appropriate operation.
  """
  status = jparse().get('status', 'positive')
  operation_state = find_op_state()
  operation_state.notify_ended(status)
  event_telem = operation_state.gen_event_telem()
  telem_man.store_event(event_telem)
  return jsonify(data=dict(success='yeah'))


def find_operation(operation_id: str) -> Operation:
  """
  Inflates (instantiates) an instance of an Operation by operation_id.
  :param operation_id: desired operation to be inflated.
  :return: Operation instance.
  """
  return Operation.inflate(operation_id)


def find_stage(operation_id, stage_id) -> Stage:
  """
  Finds the Stage with a matching operation_id and stage_id.
  :param operation_id: operation id to search by.
  :param stage_id: stage id to search by.
  :return: Stage class instance.
  """
  operation = find_operation(operation_id)
  return operation.stage(stage_id)


def find_step(operation_id, stage_id, step_id) -> Step:
  """
  Finds the Step with a matching operation_id, stage_id and step_id.
  :param operation_id: operation id to search by.
  :param stage_id: stage id to search by.
  :param step_id: step id to search by.
  :return: Step class instance.
  """
  stage = find_stage(operation_id, stage_id)
  return stage.step(step_id)


def find_op_state(**kwargs) -> OperationState:
  op_id = kwargs.get('operation_id')
  raise_on_fail = kwargs.get('raise_on_fail', True)

  token = request.headers.get('Ostid')
  op_state = OperationState.find(token) if token else None

  if not op_state:
    if utils.is_in_cluster():
      if raise_on_fail:
        raise RuntimeError('OST ID not provided in headers!')
    else:
      uuid = OperationState.gen(op_id)
      op_state = OperationState.find(uuid)

  return op_state



from flask import Blueprint, jsonify

from kama_sdk.core.core import job_client, consts
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.telem import telem_man
from kama_sdk.model.base import pure_provider_ids
from kama_sdk.model.glance import glance
from kama_sdk.model.glance.glance import Glance
from kama_sdk.model.supplier.base.provider import Provider
from kama_sdk.model.supplier.predicate import predicate
from kama_sdk.model.supplier.predicate.predicate import Predicate

controller = Blueprint('app_controller', __name__)

BASE_PATH = '/api/app'


@controller.route(f'{BASE_PATH}/compute-status', methods=['GET', 'POST'])
def run_status_compute():
  computer: Predicate = Predicate.inflate(predicate.APP_STATUS_PRED_ID)
  positive = computer.resolve()
  status = consts.rng if positive else consts.brk
  config_man.write_application_status(status)
  return jsonify(app_status=status)


@controller.route(f'{BASE_PATH}/sync-telem-hub', methods=['POST'])
def sync_telem_hub():
  job_id = job_client.enqueue_func(telem_man.upload_all_meta)
  return jsonify(job_id=job_id)


@controller.route(f'{BASE_PATH}/sync-status-hub', methods=['POST'])
def sync_status_hub():
  success = telem_man.upload_status()
  return jsonify(success=success)


@controller.route(f'{BASE_PATH}/uninstall-spec')
def uninstall_victims():
  spec_id = pure_provider_ids.deletion_spec_id
  provider: Provider = Provider.inflate(spec_id)
  spec = provider.resolve() if provider else None
  return jsonify(data=spec)


@controller.route(f'{BASE_PATH}/master_config')
def get_master_config():
  return jsonify(data=config_man.serialize())


@controller.route(f'{BASE_PATH}/deletion_selectors')
def deletion_selectors():
  deletion_map = 3
  return jsonify(data=deletion_map)


@controller.route(f'{BASE_PATH}/jobs/<job_id>/status')
def job_progress(job_id):
  status_wrapper = job_client.job_status(job_id)
  return jsonify(
    data=dict(
      status=status_wrapper.status(),
      progress=status_wrapper.progress_bundle,
      result=status_wrapper.result
    )
  )


@controller.route(f'{BASE_PATH}/glance-ids', methods=["GET"])
def glance_ids():
  """
  Returns a list of application endpoint adapter.
  :return: list of serialized adapter.
  """
  provider_id = glance.master_glances_supplier_id
  glance_descriptors = Provider.inflate(provider_id).resolve()
  glances = Glance.inflate_many(glance_descriptors, safely=True)
  serialized = list(map(Glance.fast_serialize, glances))
  return jsonify(data=serialized)


@controller.route(f'{BASE_PATH}/glances/<glance_id>', methods=["GET"])
def get_glance(glance_id: str):
  glance = Glance.inflate(glance_id)
  return jsonify(data=glance.compute_and_serialize())



import json

from flask import request

from kama_sdk.core.core import utils


def jparse():
  if utils.is_in_cluster():
    payload_str = request.data.decode('unicode-escape')
    truncated = payload_str[1:len(payload_str) - 1]
    as_dict = json.loads(truncated)
    return utils.unmuck_primitives(as_dict)
  else:
    return utils.unmuck_primitives(request.json)



from typing import Any, Optional
from werkzeug.utils import cached_property
from kama_sdk.core.core.config_man import config_man
from kama_sdk.model.supplier.base.supplier import Supplier


class MasterConfigSupplier(Supplier):

  @cached_property
  def field_key(self) -> str:
    return self.get_prop(FIELD_KEY) or self.source_data()

  @cached_property
  def reload_cmap(self) -> Optional[bool]:
    return self.get_prop(RELOAD_CMAP_KEY, None)

  def _compute(self) -> Any:
    if self.field_key == 'ns':
      return config_man.ns()
    else:
      return config_man.read_infer_type(
        self.field_key,
        reload=self.reload_cmap
      )


FIELD_KEY = 'field_key'
RELOAD_CMAP_KEY = 'reload'



from flask import Blueprint, jsonify

from kama_sdk.core.core import updates_man, job_client

controller = Blueprint('injections_controller', __name__)

BASE_PATH = '/api/injections'

@controller.route(f'{BASE_PATH}/check_newer')
def check_newer():
  new_available = not updates_man.is_using_latest_injection()
  return jsonify(data=new_available)


@controller.route(f'{BASE_PATH}/latest/preview')
def preview():
  bundle = updates_man.latest_injection_bundle()
  preview_data = updates_man.preview_injection(bundle)
  return jsonify(data=preview_data)


@controller.route(f'{BASE_PATH}/latest/apply', methods=['POST'])
def apply_latest_injection():
  action_kod = 'sdk.action.apply_latest_vendor_injections'
  job_id = job_client.enqueue_action(action_kod)
  return jsonify(data=dict(job_id=job_id))



import random
import string
from functools import reduce
from typing import Any

from werkzeug.utils import cached_property

from kama_sdk.model.supplier.base.supplier import Supplier


class RandomStringSupplier(Supplier):

  SYMBOLS_KEY = 'symbols'
  LENGTH_KEY = 'length'

  @cached_property
  def symbol_groups(self):
    return self.get_prop(self.SYMBOLS_KEY, ['letters'])

  @cached_property
  def string_length(self):
    return self.get_prop(self.LENGTH_KEY, 16)

  @cached_property
  def output_format(self):
    return None

  def _compute(self) -> Any:
    reducer = lambda whole, group: f"{whole}{self.populate_symbols(group)}"
    symbols = list(reduce(reducer, self.symbol_groups, ''))
    random.shuffle(symbols)
    scrambled = random.choices(symbols, k=self.string_length)
    return ''.join(scrambled)

  @staticmethod
  def populate_symbols(group_name: str) -> str:
    if group_name == 'letters':
      return string.ascii_letters
    elif group_name == 'numbers':
      return string.digits
    elif group_name == 'punctuation':
      return string.punctuation
    else:
      print(f"[kama_sdk:rss] unknown symbol group {group_name}")
      return ''



from typing import Tuple, Optional

from werkzeug.utils import cached_property

from kama_sdk.core.core.types import EndpointDict
from kama_sdk.core.core.utils import pwar
from kama_sdk.model.glance import graphic_and_text_glance as prt_cls, glance
from kama_sdk.model.glance.graphic_and_text_glance import GraphicAndTextGlance


class EndpointGlance(GraphicAndTextGlance):

  @cached_property
  def title(self):
    return super(EndpointGlance, self).title or 'Webpage'

  @cached_property
  def line_two(self):
    if explicit := super(EndpointGlance, self).line_two:
      return explicit
    else:
      url = self.endpoint.get('url') if self.endpoint else None
      return url or 'No URL'

  @cached_property
  def line_three(self):
    if explicit := super(EndpointGlance, self).line_three:
      return explicit
    else:
      _type = self.endpoint.get('svc_type') if self.endpoint else None
      return _type or 'Type Unknown'

  @cached_property
  def url_intent(self) -> str:
    if explicit := super(EndpointGlance, self).url_intent:
      return explicit
    elif self.endpoint and self.endpoint.get('url'):
      return f"http://{self.endpoint.get('url')}"

  @cached_property
  def endpoint(self) -> Optional[EndpointDict]:
    try:
      return self.get_prop(ENDPOINT_DATA_KEY)
    except:
      pwar(self, "endpoint load failed", True)
      return None

  @cached_property
  def legend_type(self) -> str:
    if explicit := self.config.get(glance.LEGEND_TYPE_KEY):
      return explicit
    else:
      if self.is_working():
        return 'default'
      else:
        return 'status'

  @cached_property
  def legend_icon(self) -> str:
    if explicit := super(EndpointGlance, self).legend_icon:
      return explicit
    else:
      if self.is_working():
        return self.get_prop(ICON_CONNECTED_KEY, 'open_in_new')
      # else:
      #   return self.get_prop(ICON_NOT_CONNECTED_KEY, 'error_outline')

  @cached_property
  def legend_text(self) -> str:
    if explicit := super(EndpointGlance, self).legend_text:
      return explicit
    else:
      return 'Launch' if self.is_working() else 'Site Offline'

  @cached_property
  def legend_emotion(self) -> str:
    if explicit := self.config.get(glance.LEGEND_EMOTION_KEY):
      return explicit
    else:
      return 'primaryColor' if self.is_working() else 'warning2'

  def is_working(self) -> bool:
    if data := self.endpoint:
      return data.get('url') is not None
    else:
      return False

  @cached_property
  def port_forward_spec(self):
    if data := self.endpoint:
      return data.get('port_forward_spec')
    else:
      return None

  def graphic_and_type(self) -> Tuple[str, str]:
    exp_graphic = self.config.get(prt_cls.GRAPHIC_KEY)
    exp_graphic_type = self.config.get(prt_cls.GRAPHIC_TYPE_KEY)

    if exp_graphic:
      return exp_graphic, exp_graphic_type or 'image'
    else:
      return 'language', 'icon'

  @cached_property
  def graphic_type(self):
    return self.graphic_and_type()[1]

  @cached_property
  def graphic(self):
    return self.graphic_and_type()[0]

  def content_spec(self):
    self.patch({'endpoint': self.endpoint})
    return super(EndpointGlance, self).content_spec()


ENDPOINT_DATA_KEY = 'endpoint'
SITE_LOGO_KEY = 'site_logo'
ICON_CONNECTED_KEY = 'icon_not_connected'
ICON_NOT_CONNECTED_KEY = 'icon_not_connected'



from kama_sdk.model.operation.operation import Operation
from kama_sdk.model.operation.stage import Stage


def ser_standard(operation: Operation):
  """
  Standard serializer for an Operation.
  :param operation: Operation instance.
  :return: serialized Operation dict.
  """
  return dict(
    id=operation.id(),
    title=operation.title,
    info=operation.info,
    tags=operation.tags,
    synopsis=operation.synopsis,
    stages=list(map(ser_embedded_stage, operation.stages))
  )


def ser_embedded_stage(stage: Stage):
  """
  Standard serializer for a Stage.
  :param stage: Stage instance.
  :return: serialized Stage dict.
  """
  return dict(
    id=stage.id(),
    title=stage.title,
    description=stage.info,
    first_step_id=stage.first_step_id()
  )


def ser_full(operation: Operation):
  """
  Full serializer for an Operation - includes the Operation itself as well as
  related Stages and Prerequisites.
  :param operation: Operation instance.
  :return: serialized Operation dict.
  """
  # stage_dicts = list(map(ser_embedded_stage, operation.stages))

  return dict(
    **ser_standard(operation)
    # stages=stage_dicts
  )



from flask import Blueprint, jsonify

from kama_sdk.core.core import job_client
from kama_sdk.model.action.ext.misc.run_predicates_action import RunPredicatesAction
from kama_sdk.serializers import system_test_serializer

controller = Blueprint('system_checks', __name__)

BASE_PATH = '/api/system_checks'

@controller.route(BASE_PATH)
def index():
  models = RunPredicatesAction.from_provider()
  serializer = system_test_serializer.serialize_std
  serialized = list(map(serializer, models))
  return dict(data=serialized)


@controller.route(f"{BASE_PATH}/<_id>/run", methods=['POST'])
def run(_id: str):
  if RunPredicatesAction.inflate(_id, safely=True):
    job_id = job_client.enqueue_action(_id)
    return jsonify(status='running', job_id=job_id)
  else:
    return jsonify(error=f"test not found for id='{_id}'"), 404



from typing import Dict

from flask import Blueprint, jsonify
from k8kat.auth.kube_broker import broker

from kama_sdk.core.core import utils
from kama_sdk.core.telem import telem_man
from kama_sdk.model.base.model import models_man, default_descriptors, configs_for_kinds, Model

from kama_sdk.controllers.ctrl_utils import jparse
from kama_sdk.core.core.config_man import config_man
from kama_sdk.core.ktea.ktea_provider import ktea_client
from kama_sdk.core.telem.telem_perms import TelemPerms

controller = Blueprint('status_controller', __name__)


@controller.route('/api/ping')
def ping():
  return jsonify(ping='pong')


@controller.route('/api/status/telem-perms')
def status_telem_perms():
  """
  Returns the default user perms.
  :return: serialized user perms.
  """
  raw_perms: Dict = TelemPerms().user_perms()
  return jsonify(data=raw_perms)


@controller.route('/api/status/telem-perms', methods=['POST'])
def status_patch_telem_perms():
  """
  Patches the user perms.
  :return: patched user perms.
  """
  patch_values: Dict = jparse()['data']
  TelemPerms().patch(patch_values)
  return jsonify(data=TelemPerms().user_perms())

@controller.route('/api/status/telem')
def telem_status():
  return jsonify(
    is_mongo_storage_ready=telem_man.is_storage_ready()
  )

@controller.route('/api/status')
def status():
  """
  Checks Wiz's status.
  :return: dict containing status details.
  """

  if not is_healthy():
    broker.connect()

  return jsonify(
    sanity='2',
    app_id=config_man.app_id(),
    nectwiz_env=utils.run_env(),
    is_training_mode=config_man.is_training_mode(),
    is_healthy=is_healthy(),
    install_id=config_man.install_id(),
    install_token=config_man.install_token(),
    cluster_connection=dict(
      is_k8kat_connected=broker.is_connected,
      connect_config=broker.connect_config
    ),
    ns=config_man.ns(),
    ktea_config=config_man.ktea_desc(),
    wiz_config=config_man.kama_desc(),
    ktea_defaults=config_man.default_vars(),
    ktea_variables=config_man.user_vars()
  )


@controller.route('/api/status/templated_manifest')
def templated_manifest():
  return jsonify(
    data=ktea_client().template_manifest()
  )


@controller.route('/api/status/descriptors')
def dump_descriptors():
  descriptors = models_man.descriptors()
  return jsonify(data=descriptors)


@controller.route('/api/status/descriptors/<kind>')
def dump_descriptors_by_kind(kind):
  master = Model.inflate(kind)
  cls_pool = master.lteq_classes(models_man.classes())
  descriptors = configs_for_kinds(models_man.descriptors(), cls_pool)
  return jsonify(data=descriptors)


@controller.route('/api/status/descriptors/<kind>/<res_id>')
def dump_descriptor(kind, res_id):
  model_class: Model = Model.kind2cls(kind)
  instance = model_class.inflate(res_id)
  return jsonify(data=instance.to_dict())


@controller.route('/api/status/default-descriptors')
def dump_default_descriptors():
  descriptors = default_descriptors()
  return jsonify(data=descriptors)


def is_healthy() -> bool:
  if broker.is_connected:
    return config_man.load_master_cmap() is not None
  else:
    return False



from typing import Dict, List
from urllib.parse import quote_plus

import requests

from kama_sdk.core.core.types import K8sResDict
from kama_sdk.core.ktea.ktea_client import KteaClient


class HttpKteaClient(KteaClient):

  def load_default_values(self) -> Dict[str, str]:
    return self.http_get(f"/values?{self.any_cmd_args()}")

  def template_manifest(self, values: Dict) -> List[K8sResDict]:
    endpoint = f"/template?{self.template_cmd_args()}"
    return self.http_post(endpoint, values)

  def base_url(self):
    ktea = self.ktea
    version = ktea.get('version')
    version_part = f"/{version}" if version else ''
    return f"{ktea['uri']}{version_part}"

  def http_post(self, endpoint, payload):
    url = f'{self.base_url()}{endpoint}'
    response = requests.post(url, json=payload)
    return response.json().get('data')

  def http_get(self, endpoint):
    url = f'{self.base_url()}{endpoint}'
    response = requests.get(url)
    return response.json().get('data')

  def any_cmd_args(self) -> str:
    args_str = super().any_cmd_args()
    if args_str:
      return f"args={quote_plus(args_str)}"
    return ''

  def template_cmd_args(self, *args) -> str:
    std_part = self.any_cmd_args()
    if not self.release_name():
      print("[kama_sdk:ktea+client] fatal ns is blank")
    ns_part = f"release_name={self.release_name()}"
    std_part = f"&{std_part}" if std_part else ""
    return f"{ns_part}{std_part}"



from typing import List

from k8kat.res.base.kat_res import KatRes
from k8kat.res.pod.kat_pod import KatPod
from werkzeug.utils import cached_property

from kama_sdk.core.core.types import PodStatusSummary
from kama_sdk.model.supplier.base.supplier import Supplier


class PodStatusSummariesSupplier(Supplier):

  @cached_property
  def workload(self) -> KatRes:
    return self.source_data()

  @cached_property
  def pods(self) -> List[KatPod]:
    return self.workload.pods()

  def resolve(self) -> List[PodStatusSummary]:
    return list(map(pod_mini_status, self.pods))


def pod_mini_status(pod: KatPod) -> PodStatusSummary:
  return PodStatusSummary(
    ternary_status=pod.ternary_status(),
    pod_name=pod.name,
    phase=pod.phase
  )



import re
from typing import Callable

REGEX = r"\${(.*?)}"

def interpolate(string: str, callback: Callable) -> str:
  new_string = string
  matches = list(re.finditer(REGEX, string))
  for match in matches:
    expr_with_characters = match.group()
    expr = re.findall(REGEX, expr_with_characters)[0]
    interpolated_str = callback(expr)
    new_string = new_string.replace(expr_with_characters, interpolated_str)
  return new_string



from typing import Optional, Any

from kama_sdk.model.input.generic_input import GenericInput


class CheckboxesInput(GenericInput):
  pass

class CheckboxInput(GenericInput):
  def compute_inferred_default(self) -> Optional[Any]:
    return False


class SelectInput(GenericInput):
  pass



import subprocess
from typing import Dict, List

import yaml

from kama_sdk.core.core.types import K8sResDict, KteaDict

from kama_sdk.core.ktea.ktea_client import KteaClient

tmp_vars_path = '/tmp/necwiz-tmp-values.yaml'


class LocalExecKteaClient(KteaClient):
  def load_default_values(self) -> Dict[str, str]:
    raw = exec_cmd(self.ktea, f"show values . {self.any_cmd_args()}")
    return yaml.load(raw, Loader=yaml.FullLoader)

  def template_manifest(self, values) -> List[K8sResDict]:
    write_values_to_tmpfile(values)
    cmd_args = self.template_cmd_args(tmp_vars_path)
    raw = exec_cmd(self.ktea, f"template {cmd_args}")
    return list(yaml.load_all(raw, Loader=yaml.FullLoader))


def write_values_to_tmpfile(values: Dict):
  file_content = yaml.dump(values)
  with open(tmp_vars_path, 'w') as file:
    file.write(file_content)


def exec_cmd(ktea: KteaDict, cmd):
  uri, ver = ktea['uri'], ktea.get('version')
  exec_name = f"{uri}:{ver}" if ver else uri
  full_cmd = f"{exec_name} {cmd}".split(" ")
  output = subprocess.check_output(full_cmd).decode('utf-8')
  return output



import base64
import json
from typing import Dict

from kama_sdk.model.base.model import Model
from kama_sdk.model.concern.concern_attr_adapter import ConcernAttrAdapter
from kama_sdk.model.concern.concern_detail_adapter import ConcernDetailAdapter, AttrPanelDescriptor
from kama_sdk.model.concern.concern_set import ConcernSet
from kama_sdk.model.concern.concern_table import ConcernTable

def encode_seed(seed_dict: Dict) -> str:
  utf_json = json.dumps(seed_dict)
  return base64.b64encode(utf_json)


def ser_simple_child(model: Model) -> Dict:
  return {
    'id': model.id(),
    'title': model.title,
    'info': model.info
  }


def ser_attr_meta(model: ConcernAttrAdapter) -> Dict:
  return {
    'label': model.label,
    'title': model.title,
    'info': model.info
  }


def ser_attr_panel(model: AttrPanelDescriptor) -> Dict:
  return {
    **ser_simple_child(model),
    'attributes': list(map(ser_attr_meta, model.attr_adapters()))
  }


def serialize_set_meta(concern_set: ConcernSet, seeds=False):
  bundle = {
    **ser_simple_child(concern_set),
    'one_shot': concern_set.is_one_shot,
    'view_type': concern_set.view_type()
  }

  serialized_cols = None
  serialized_seeds = None

  if isinstance(concern_set, ConcernTable):
    serialized_cols = concern_set.column_definitions

  if seeds and concern_set.is_one_shot:
    seeds = concern_set.compute_concern_seeds()
    serialized_seeds = list(map(encode_seed, seeds))

  bundle['columns'] = serialized_cols
  bundle['seeds'] = serialized_seeds

  return bundle


def ser_detail_meta(adapter: ConcernDetailAdapter) -> Dict:
  attr_panel_descs = adapter.attr_panel_descs()
  value_panel_descs = adapter.value_panel_descs()
  related_set_descs = adapter.related_sets()

  return {
    'id': adapter.id(),
    'title': adapter.title,
    'info': adapter.info,
    'attr_panels': list(map(ser_attr_panel, attr_panel_descs)),
    'value_panels': list(map(ser_simple_child, value_panel_descs)),
    'related_sets': list(map(serialize_set_meta, related_set_descs)),
    'actions': list(map(ser_simple_child, adapter.actions())),
    'operations': list(map(ser_simple_child, adapter.operations()))
  }



from typing import Dict, Optional, List

from kama_sdk.model.base.model import Model
from kama_sdk.model.operation.operation import Operation
from kama_sdk.model.operation.stage import Stage


class OperationRunSimulator(Model):
  def __init__(self, config: Dict):
    super().__init__(config)
    self.force_ns: str = config['namespace']
    self.operation: Operation = Operation.inflate(config['operation'])
    self.stage_sims: List[Dict] = config.get('stages', [])
    self.validate_input: bool = config.get('validations', True)
    self.run_preflight: bool = config.get('preflight', False)
    self.print_progress: bool = config.get('print_progress', True)

  def stage_at(self, index: int) -> Optional[Stage]:
    if index < len(self.stage_sims):
      stage_sim = self.stage_sims[index]
      return self.operation.stage(stage_sim['stage'])

  def step_input(self, stage_index: int, step_id: str) -> Dict:
    stage_sim = self.stage_sims[stage_index]
    step_sim_finder = lambda step_dict: step_dict['step'] == step_id
    step_sim = next(filter(step_sim_finder, stage_sim['steps']))
    return step_sim['input']



from werkzeug.utils import cached_property

from kama_sdk.model.supplier.base.supplier import Supplier


class Provider(Supplier):
  @cached_property
  def serializer_type(self) -> str:
    return 'identity'



from copy import deepcopy
from typing import Optional, Union

from kama_sdk.core.core.types import ErrCapture


class ActionError(Exception):
  err_capture: ErrCapture
  is_original: bool

  def __init__(self, err_capture: ErrCapture, is_original=None):
    err: ErrCapture = deepcopy(err_capture)
    self.err_capture = {
      **err,
      'fatal': err.get('fatal', False),
      'is_original': True if is_original is None else is_original
    }
    super().__init__(err.get('reason'))


class FatalActionError(ActionError):

  def __init__(self, err_capture: ErrCapture, is_original=None):
    new_capt = {**err_capture, 'fatal': True}
    super().__init__(new_capt, is_original)


ActionErrOrException = Optional[Union[Exception, ActionError]]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name="kama-sdk-py",
  version="0.0.006",
  author="Nectar Cloud Software",
  author_email="xavier@codenectar.com",
  description="Kubernetes Application Management API (KAMA) SDK",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/nectar-cs/kama-sdk-py",
  packages=setuptools.find_packages(exclude=[
    "kama_sdk.tests.*",
    "kama_sdk.tests"
  ]),
  package_data={
    'kama_sdk': [
      'assets/*.*'
      'model/**/yamls/**/*',
    ]
  },
  include_package_data=True,
  install_requires=[
    'flask>=1.1',
    'flask-cors',
    'k8kat>=0.0.231',
    'requests',
    'termcolor>=1.1.0',
    'pymongo',
    'cachetools>=3.1',
    'redis>=3',
    'rq',
    'validators',
    'humanize>=3.2.0',
    'jq>=1.1.2'
  ],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires='>=3.8'
)



import traceback
from typing import Dict, Any

from werkzeug.utils import cached_property

from kama_sdk.model.input.generic_input import GenericInput


class SliderInput(GenericInput):
  @cached_property
  def min(self) -> int:
    return self.get_prop('min', 0)

  @cached_property
  def max(self) -> int:
    return self.get_prop('max', 10)

  @cached_property
  def step(self) -> int:
    return self.get_prop('step', 1)

  @cached_property
  def suffix(self):
    return self.get_prop('suffix', '')

  def extras(self) -> Dict[str, any]:
    return dict(
      min=self.min,
      max=self.max,
      step=self.step,
      suffix=self.suffix
    )

  def sanitize_for_validation(self, value: Any) -> Any:
    if value is not None and self.suffix:
      # noinspection PyBroadException
      try:
        return float(str(value).split(self.suffix)[0])
      except:
        print("[kama_sdk::slider_input] sanitize input failed:")
        print(traceback.format_exc())
        return value
    else:
      return value
